#!/usr/bin/perl

use IO::File;
use File::Path;

#####################################################
# History
# 
# 2016/04/07 : (1) Add Codec.h
#              (2) Change DESC_HDR_PATH to DESC_LIB_PATH
#              (3) Change DESC_C_PATH   to DESC_OPEN_SRC_PATH
#
# 2015/11/17 :
#              (1) Fix typedef parser issue
#              (2) Split SRC_PATH to DESC_HDR_PATH and DESC_C_PATH
#
#####################################################
# Options
#####################################################

$DESC_LIB_PATH        = "..\\Source\\P2pdLib";
@DESC_LIB_FILES       = ("USB.h","USB_IP.h","UAVC.h");

$DESC_OPEN_SRC_PATH   = "..\\Source";
@DESC_OPEN_SRC_FILES  = ("USB_Descriptor.c","Codec.h");

$DESC_BIN_FILE        = "USB_DESC.bin";
$DESC_ARY_ID_FILE     = "USB_DESC_ARY_ID.h";

$LOG_FILE_NAME        = "usb_desc.log";

$disp_warn = 0;


#####################################################
# Main Function
#####################################################

main();

sub main
{

  # Open output files
  open(LOG_F,"> $LOG_FILE_NAME") or die("Can't open $LOG_FILE_NAME\n");
  open(BIN_F,"> $DESC_BIN_FILE") or die("Can't open $DESC_BIN_FILE\n");
  
  # Get information from DESC Header files
  foreach $desc_lib_file (@DESC_LIB_FILES)
  { get_c_src_file_info("${DESC_LIB_PATH}\\${desc_lib_file}",""); }
  
  # Get information from DESC C files
  foreach $desc_open_src_file (@DESC_OPEN_SRC_FILES)
  { 
  	$XFER_FILE_NAME = $desc_open_src_file;
  	$XFER_FILE_NAME =~ s/\./_xfer\./;
  	open(XFER_F,"> $XFER_FILE_NAME") or die("Can't open $XFER_FILE_NAME\n");
  	get_c_src_file_info("${DESC_OPEN_SRC_PATH}\\$desc_open_src_file","TRANSFER");
  	close XFER_F; 
  }
    
  # Output descriptor binary file
  output_desc_bin_file();
  
  # Close output file
  close LOG_F;
  close BIN_F;

  
  
}

# --------------------------------------

sub get_c_src_file_info
{

  my ($c_src_f,$in_opt) = @_;
  my ($line); 

  $cur_file = $c_src_f;
  
  print "--> get_c_src_file_info($cur_file)\n";
  
  open(C_SRC_F,"< $c_src_f") or die "Can't open $c_src_f\n";
  @C_SRC_FILE_DATA=<C_SRC_F>;
  close C_SRC_F;
  
  $cur_line = 0;
  foreach $line (@C_SRC_FILE_DATA)
  { parse_c_line($line,$in_opt); }

}

# --------------------------------------

sub parse_c_line
{
  my ($in_line,$in_opt) = @_;
  
  my ($org_line);
  
  $org_line = $in_line;
  
  #increase current line number
  $cur_line++;

  # print "$in_line";
  
  # Remove heading spaces
  $in_line =~ s/^\s+//;
  
  # Get define name after // -->   // #define DEFINE_NAME
  if($in_line =~ /\/\/\s* #define\s+(\w+)/)
  { $define_in_comment = $1; }
  else
  { $define_in_comment = ""; }
  
  # Remove comment
  $in_line = remove_c_comment($in_line);
  
  # Proc "#define"
  if($in_line =~ /#define\s+(\w+)\s+(.*)$/)
  {
    $def_name  = $1;
    $def_value = $2;
    $def_value =~ s/\s//g;
    
    if($c_def{$def_name} ne "")
    { 
      if($disp_warn==1)
      { print "Warn : Proc \"#define\" --> ($def_name) has been defined\n"; }
    }
    else
    { $c_def{$def_name} = $def_value; }
  }
  # Proc "typedef struct" start
  elsif ($in_line =~ /typedef\s+struct\s+(\w+)/)
  {
  	$struct_name = $1;
    $in_typedef_struct_block = 1;
    $typedef_struct[$typedef_struct_idx]{name} = $struct_name;
  }
  elsif ($in_typedef_struct_block==1)
  {
    # end of typedef struct block
    if($in_line =~ /}\s*(\w+)/)
    {
      $typedef_struct_inst_name = $1;
      $in_typedef_struct_block = 0;
      $typedef_struct[$typedef_struct_idx]{inst_name} = $typedef_struct_inst_name;
      
      $typedef_struct_idx++;
    }
    # U8 array
    elsif($in_line =~ /U8\s+(\w+)\[(\d+)\];/)
    { $ary_num = $2; 
      $typedef_struct[$typedef_struct_idx]{byte_size} += 1*$ary_num;
    }
    # U8 var
    elsif($in_line =~ /U8\s+(\w+);/)
    { $typedef_struct[$typedef_struct_idx]{byte_size} += 1; }
    # U16 array
    elsif($in_line =~ /U16\s+(\w+)\[(\d+)\];/)
    { $ary_num = $2; 
      $typedef_struct[$typedef_struct_idx]{byte_size} += 2*$ary_num;
    }
    # U16 var
    elsif($in_line =~ /U16\s+(\w+);/)
    { $typedef_struct[$typedef_struct_idx]{byte_size} += 2; }
    # U32 array
    elsif($in_line =~ /U32\s+(\w+)\[(\d+)\];/)
    { $ary_num = $2; 
      $typedef_struct[$typedef_struct_idx]{byte_size} += 4*$ary_num;
    }
    # U32 var
    elsif($in_line =~ /U32\s+(\w+);/)
    { $typedef_struct[$typedef_struct_idx]{byte_size} += 4; }
    elsif($in_line =~ /\{/) {}
    else
    { die("Unknown format in typedef struct block ==> $cur_file($cur_line), line=$in_line"); }
  }
  # Proc "typedef struct" end
  
  # Proc "Ux code []" start
  elsif($in_line =~ /^U(\d+)\s+code\s+(\w+)\[/)
  {
  	$unit_size = $1/8;
  	$code_ary_name = $2;
    $in_ux_code_block = 1;
    
    # Check if ARY name exist
    if(get_ary_value($code_ary_name,"idx") eq "CAN_NOT_FIND_ARY")
    { 
      $ux_code_block_exist = 0; 
      $code_ary[$code_ary_idx]{unit_size} = $unit_size;
      $code_ary[$code_ary_idx]{name}      = $code_ary_name;
      
      if($define_in_comment ne "")
      { $code_ary[$code_ary_idx]{section_id_define} = $define_in_comment; 
      
        if($c_def{$define_in_comment} ne "")
        { 
          # @DESC_ARYs data format
          # "$ary_item"         :  assign array item name, Section ID = previous Section ID + 1
          # "$ary_item->$num"   :  assign array item name, Section ID = $num
	
        	push (@DESC_ARYS,${code_ary_name}."->".$c_def{$define_in_comment}); 
        }
        else
        { die("Error : Proc \"Ux code []\" --> can't get define value for ($define_in_comment)\n"); }
        
      }
      
      print LOG_F "\n=====================================\n";
      print LOG_F "$code_ary_name --> $unit_size\n";
    }
    else
    { 
      $ux_code_block_exist = 1; 
      if($disp_warn==1)
      {
        print LOG_F "\n=====================================\n";
        print LOG_F "Warn : code_ary($code_ary_name) exist\n";
      }
    }
    
    # Check end of code_ary    
    if($in_line =~ /};/)
    { $in_ux_code_block  = 0; 
    	$pre_ux_code_block = 1;
    	
    	if($ux_code_block_exist==0)
    	{
    	  $code_ary[$code_ary_idx]{contain} = $in_line;
        proc_code_ary($code_ary_idx);
        if($in_opt eq "TRANSFER")
        { print XFER_F $code_ary[$code_ary_idx]{transfer}; }
        $code_ary_idx++;	
      }
    }
    
    if($ux_code_block_exist==0)    
    { $code_ary[$code_ary_idx]{contain} = $in_line; }
    
  }
  elsif($in_ux_code_block==1)
  {
    if($ux_code_block_exist==0) 
    { $code_ary[$code_ary_idx]{contain} .= $in_line;	 }
    
    # Check end of cod_ary
    if($in_line =~ /};/)
    { 
      $in_ux_code_block = 0; 
      if($ux_code_block_exist==0) 
      {
        proc_code_ary($code_ary_idx);
        if($in_opt eq "TRANSFER")
        { print XFER_F $code_ary[$code_ary_idx]{transfer}; }
      
        $code_ary_idx++;
      }
    }
  }
  # Proc "Ux code []" end
  
  # output transfer file
  if($in_opt eq "TRANSFER")
  {
    if($in_ux_code_block==1 || $pre_ux_code_block==1)
    {}
    else
    { print XFER_F $org_line; }
  
  }
  
  $pre_ux_code_block = $in_ux_code_block;
  
}

# --------------------------------------
sub remove_c_comment
{
  
  my ($in_line) = @_;
  
  # Remove enter char
  $in_line =~ s/[\r\n]//g;
  
  # Remove comment
  $in_line =~ s/\/\/.*$//;
  
  if($in_line =~ s/\/\*.*$//)
  { $comment_line = 1; }
  elsif($in_line =~ s/^\*\///)
  { $comment_line = 0; }
  elsif($comment_line == 1)
  { return "SKIP_FOR_COMMENT"; }
  
  return $in_line;
    
}

# --------------------------------------

sub proc_code_ary
{
  my ($in_code_ary_idx) = @_;
  
  my ($cur_unit_size, $cur_code_ary_contain);
  
  $cur_unit_size        = $code_ary[$in_code_ary_idx]{unit_size};
  $cur_code_ary_contain = $code_ary[$in_code_ary_idx]{contain};
  
  # Remove spaces
  $cur_code_ary_contain =~ s/{\s+/{/g;
  $cur_code_ary_contain =~ s/\s+}/}/g;
  $cur_code_ary_contain =~ s/\s+\,/,/g;
  $cur_code_ary_contain =~ s/\,\s+/,/g;
  
  # Extract {...}
  $cur_code_ary_contain =~ /(.*){(.*)}/;
  $cur_code_ary_declare = $1;
  $cur_code_ary_contain = $2;
  
  # Assign array declare
  $code_ary[$in_code_ary_idx]{transfer} = $cur_code_ary_declare."{";
  
  # Extract array item
  @cur_code_ary_contain_item = split(",",$cur_code_ary_contain);
   
  print LOG_F "--> Proc $code_ary[$in_code_ary_idx]{name}\n";
  foreach $item (@cur_code_ary_contain_item)
  {
    
    # print "$item\n";
    
    if($item =~ /'([\w\s])'/)
    {
      $char_value = ord($1);
      ($code,$xfer) = format_item($cur_unit_size,$char_value);
      $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
    }
    # Proc sizeof(struct
    elsif($item =~ /sizeof\(struct\s+(\w+)/)
    {
      $struct_name = $1;
      $found=0;
      for($idx=0;$idx<$typedef_struct_idx;$idx++)
      {
        # print "$typedef_struct[$idx]{name}\n";
        if($typedef_struct[$idx]{name} eq $struct_name)
        {
          $found=1;
          if($cur_unit_size,$typedef_struct[$idx]{byte_size}==0)
          {die ("Error : Proc sizeof(struct --> found ($struct_name) byte_size=0\n");}
          else
          { ($code,$xfer) = format_item($cur_unit_size,$typedef_struct[$idx]{byte_size}); }
          $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
          break;
        }
      }
      if($found==0)
      { die("Error : Proc sizeof(struct --> can't find related struct($struct_name)\n"); }
    }
    # Proc sizeof(ARY)
    elsif($item =~ /sizeof\((\w+)/)
    {
      $ary_name = $1;
      ($code,$xfer) = format_item($cur_unit_size, get_ary_value($ary_name,"byte_size"));
      $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
    }
    # Proc 0x...
    elsif($item =~ /0x(\w+)/)
    {
      $hex_value = hex($1);
      ($code,$xfer) = format_item($cur_unit_size, $hex_value);
      $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
    }
    # Proc DEC value
    elsif($item =~ /^(\d+)$/)
    {
      $dec_value = $1;
      ($code,$xfer) = format_item($cur_unit_size, $dec_value);
      $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
    }
    # Proc W2B(DEF) or DW2B(DEF)
    elsif($item =~ /W2B\((\w+)/)
    {
      $cur_def = $1;
      
      if($cur_unit_size != 1)
      { die("Error : Proc W2B(DEF) --> cur_unit_size($cur_unit_size) != 1\n"); }

      $cur_def_value = get_def_value($cur_def);
      
      # Byte0
      ($tmp_code,$tmp_xfer) = format_item($cur_unit_size, ($cur_def_value>>0) & 0xFF);
      $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
      $code = $tmp_code;
      $xfer = $tmp_xfer;
      
      # Byte1
      ($tmp_code,$tmp_xfer) = format_item($cur_unit_size, ($cur_def_value>>8) & 0xFF);
      $code .= $tmp_code;
      $xfer .= ",".$tmp_xfer;
      $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
      
      if($item =~ /DW2B/)
      {
        # Byte2
        ($tmp_code,$tmp_xfer) = format_item($cur_unit_size, ($cur_def_value>>16) & 0xFF);
        $code .= $tmp_code;
        $xfer .= ",".$tmp_xfer;
        $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
        
        # Byte3
        ($tmp_code,$tmp_xfer) = format_item($cur_unit_size, ($cur_def_value>>24) & 0xFF);
        $code .= $tmp_code;
        $xfer .= ",".$tmp_xfer;
        $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size;
      }
    }
    # Proc DEF or ARY
    elsif($item =~ /^(\w+)$/)
    {
      $cur_def_or_ary = $1;
      $cur_def_value = get_def_value($cur_def_or_ary);
      
      if ($cur_def_value eq "CAN_NOT_FIND_DEF")
      {
        $cur_ary_value = get_ary_value($cur_def_or_ary,"value");
        
        if($cur_ary_value eq "CAN_NOT_FIND_ARY")
        {
          die("Error : Proc DEF or ARY --> Can't find ($cur_def_or_ary)\n");
        }
        else
        {
          $code = $cur_ary_value;
          $code_ary[$in_code_ary_idx]{byte_size}+= (length($cur_ary_value)/2);
          
          $xfer = $cur_ary_value;
          if($cur_unit_size==1)
          { $xfer =~ s/(\w\w)/,0x$1/g; }
          elsif($cur_unit_size==2)
          { $xfer =~ s/(\w\w\w\w)/,0x$1/g; }
          elsif($cur_unit_size==4)
          { $xfer =~ s/(\w\w\w\w\w\w\w\w)/,0x$1/g; }
          else
          { die("Error : Proc DEF or ARY --> Un-known cur_unit_size($cur_unit_size)\n"); }
          $xfer =~ s/^,//;
          
        }
      }
      else
      { ($code,$xfer) = format_item($cur_unit_size, $cur_def_value); 
        $code_ary[$in_code_ary_idx]{byte_size}+=$cur_unit_size; 
      }
    }
    else
    { $code = "???"; }
    
    print LOG_F "$item --> $code\n";
    
    $code_ary[$in_code_ary_idx]{value}     .= $code;
    $code_ary[$in_code_ary_idx]{transfer}  .= ",".$xfer;
    
  }
  
  
  # print LOG_F "$code_ary[$in_code_ary_idx]{name} = @cur_code_ary_contain_item\n";

  $code_ary[$in_code_ary_idx]{transfer} .= "};\n";
  $code_ary[$in_code_ary_idx]{transfer} =~ s/{\,/{/;
  	
}

# --------------------------------------

sub format_item
{
  my ($cur_unit_size, $cur_value) = @_;
  
  my ($value_str);
  
  if($cur_unit_size==1)
  { $value_str = sprintf ("%2.2X", $cur_value); }
  elsif($cur_unit_size==2)
  {  $value_str = sprintf ("%4.4X", $cur_value); }
  elsif($cur_unit_size==4)
  {  $value_str = sprintf ("%8.8X", $cur_value); }
  else
  { die("Error : format_item() --> Unknown cur_unit_size($cur_unit_size)\n"); }
  
  return ($value_str, "0x$value_str");
  
}

# --------------------------------------
sub get_def_value
{
	my ($cur_def) = @_;
	
	my ($cur_def_value,$op1,$op2,$ary_name);
	
  $cur_def_value = $c_def{$cur_def};
  
  # print "get_def_value($cur_def)\n";
  
  # op1+op2
  if($cur_def_value =~ /^(\w+)\+(\w+)$/)
  { 
    $op1 = $1;
    $op2 = $2;
    $cur_def_value = get_def_value($op1) + get_def_value($op1);
  }
  # sizeof(ARY)
  elsif($cur_def_value =~ /sizeof\((\w+)/)
  {
    $ary_name = $1;
    return get_ary_value($ary_name,"byte_size");
  }
  # CAN_NOT_FIND_DEF
  elsif($cur_def_value eq "")
  {
  	if($disp_warn == 1)
  	{ print ("Warn : get_def_value($cur_def) --> Empty define for ($cur_def)\n");  }
    $cur_def_value = "CAN_NOT_FIND_DEF";
  }
  # Hex value
  elsif($cur_def_value =~ s/0x(\w+)//)
  { $cur_def_value = hex($1); }
  # Dec value
  elsif ($cur_def_value =~ /^(\d+)$/)
  { $cur_def_value = $1; }
  elsif($cur_def_value =~ /^(\w+)$/)
  { $cur_def_value = get_def_value($cur_def_value); }
  # Unknown format
  else
  { die ("Error : get_def_value() --> Unknown format for def_value($cur_def_value)\n"); }
  
  return $cur_def_value;
}

# --------------------------------------
sub get_ary_value
{
  my ($cur_ary,$cur_key) = @_;
  
  my ($cur_ary_value,$found,$idx);

  $found=0;
  for($idx=0;$idx<$code_ary_idx;$idx++)
  {
    # print "$code_ary[$idx]{name}\n";
    if($code_ary[$idx]{name} eq $cur_ary)
    {
      $found=1;
      if($cur_key eq "byte_size")
      {
        if($code_ary[$idx]{byte_size} == 0)
        { die("Error : get_ary_value($cur_ary,$cur_key) --> byte_size=0\n"); }
        return  ($code_ary[$idx]{byte_size});
      }
      elsif($cur_key eq "value")
      {
        if($code_ary[$idx]{value} eq "")
        { die("Error : get_ary_value($cur_ary,$cur_key) --> value is empty\n"); }
        return $code_ary[$idx]{value};
      }
      elsif($cur_key eq "section_id_define")
      {
        if($code_ary[$idx]{section_id_define} eq "")
        { die("Error : get_ary_value($cur_ary,$cur_key) --> value is empty\n"); }
        return $code_ary[$idx]{section_id_define};
      }
      elsif($cur_key eq "idx")
      {
        return $idx;
      }
      else
      { die("Error : get_ary_value($cur_ary,$cur_key) --> Un-support key\n"); }
      break;
    }
  }
  if($found==0)
  { return "CAN_NOT_FIND_ARY"; }
  
}

# --------------------------------------

sub output_desc_bin_file
{
  my ($ARY_HDR_ptr, $ARY_DATA_ptr,$ARY_ID,  $idx, $ary_byte_size,$ary_contain);
  
  my (@OUT_BIN_ARY);
  
  # open ARY ID file
  open(ID_F,">$DESC_ARY_ID_FILE") or &err("Cannot open file $DESC_ARY_ID_FILE due to $!");
  
  print ID_F <<EOF;
/*=========================================
  ARRAY ID File
  =========================================*/
  
EOF
  
  
  $ARY_HDR_ptr   = 0;
  $ARY_DATA_ptr  = 256;
  $ARY_ID        = 0;
  
  foreach $cur_ary_item (@DESC_ARYS)
  {
    
    if($cur_ary_item =~ /^(\w+)->(\d+)$/)
    {
      $cur_ary = $1;
      $ARY_ID  = $2;
      
      # Check if $ARY_ID has been used
      check_ARY_ID_exist($ARY_ID,$cur_ary);
    }
    elsif($cur_ary_item =~ /^(\w+)$/)
    {
      $cur_ary = $1;
      
      # Check if $ARY_ID has been used
      check_ARY_ID_exist($ARY_ID,$cur_ary);
    }
    else
    { die("Error : output_desc_bin_file() --> Unknown cur_ary_item format -> $cur_ary_item\n"); }
    
    # output array ID to ARY_ID file
    $cur_section_id_define = get_ary_value($cur_ary,"section_id_define");
    printf ID_F "#define %-30.30s %-10.10s // %s\n", "ID_".$cur_ary, $ARY_ID, $cur_section_id_define;
    
    # Get cur_ary byte size
    $ary_byte_size = get_ary_value($cur_ary,"byte_size");
    # printf "ary_byte_size=%2.2x\n",$ary_byte_size;
    
    if($ary_byte_size eq "CAN_NOT_FIND_ARY")
    { die("Error : output_desc_bin_file() --> Can't find array($cur_ary)\n"); }
  
    # Set array header info
    # -------------------------------
    #   Byte 0   : ID
    #   Byte 1~3 : Reserved
    #   Byte 4   : ADDR[0]
    #   Byte 5   : ADDR[1]
    #   Byte 6   : Length[0]
    #   Byte 7   : Length[1]
    # -------------------------------
    
    # ID
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", $ARY_ID);                    $ARY_HDR_ptr++;
    # Reserved                                                                 
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", 0);                          $ARY_HDR_ptr++;
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", 0);                          $ARY_HDR_ptr++;
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", 0);                          $ARY_HDR_ptr++;
    # ADDR
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", ($ARY_DATA_ptr>>0) & 0xFF);  $ARY_HDR_ptr++;
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", ($ARY_DATA_ptr>>8) & 0xFF);  $ARY_HDR_ptr++;
    # Length(byte)
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", ($ary_byte_size>>0) & 0xFF); $ARY_HDR_ptr++;
    $OUT_BIN_ARY[$ARY_HDR_ptr] = sprintf("%2.2X", ($ary_byte_size>>8) & 0xFF); $ARY_HDR_ptr++;

    # Increase ARY_ID
    $ARY_ID++;

    # Set array data value
    $ary_value = get_ary_value($cur_ary,"value");
    while ($ary_value =~ s/^(\w\w)//)
    { $OUT_BIN_ARY[$ARY_DATA_ptr] = $1; 
      $ARY_DATA_ptr++;
    }
    
    # Added 4 FF as array end
    $OUT_BIN_ARY[$ARY_DATA_ptr] = "FF"; $ARY_DATA_ptr++;
    $OUT_BIN_ARY[$ARY_DATA_ptr] = "FF"; $ARY_DATA_ptr++;
    $OUT_BIN_ARY[$ARY_DATA_ptr] = "FF"; $ARY_DATA_ptr++;
    $OUT_BIN_ARY[$ARY_DATA_ptr] = "FF"; $ARY_DATA_ptr++;
    
  }

  # output to BIN file
  
  open(BIN_F,">$DESC_BIN_FILE") or &err("Cannot open file $DESC_BIN_FILE due to $!");
  binmode BIN_F;

  for($idx=0;$idx<$ARY_DATA_ptr;$idx++)
  {
    $byte_data = $OUT_BIN_ARY[$idx];
    print BIN_F pack("H2",$byte_data);
  }
  
  close BIN_F;
  close ID_F;
  
  print "==================================\n";
  print " !!! Gen $DESC_BIN_FILE done !!!  \n";
  print "==================================\n";
  
}

# --------------------------------------

sub check_ARY_ID_exist
{
	my($in_id,$in_name) = @_;
  
  my($idx);
  
  for($idx=0;$idx<=$#ARY_IDs;$idx++)
  { 
    if($in_id == $ARY_IDs[$idx]) 
    { 
    	die("Error : check_ARY_ID_exist() in_id=$in_id has been defined for $ARY_ID_NAMEs[$idx]\n"); 
    }
  }
 
  push(@ARY_IDs,      $in_id); 
  push(@ARY_ID_NAMEs, $in_name);  
  
}

